using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;

namespace InfoVis.Tree
{
    /**
     * Class representing a Radial Treelayout
     */
    public class RadialTreeLayout : TreeLayout
    {
        #region Constructor
        /**
         * CTR of a RadialTreeLayout
         * @param m_Tree Tree where the layout should be built
         */
        public RadialTreeLayout(ref Tree m_Tree)
        {
            mTree = m_Tree;
            mLinesSelected = false;
        }
        #endregion

        #region Methods
        /**
         * computes the layout of all subnodes of the current Node
         * @param the SubNode where the layout depends on
         * @param angleMin the min absolute angle of circlesector
         * @param angleMax the max absolute angle of circlesector
         * @param layer the current layer in the tree
         */
        private void computeSubNodeLayout(TreeNode subNode, float angleMin, float angleMax, int layer)
        {
            float step = (angleMax - angleMin) / ((subNode.mIsLeafContent) ? (2) : (subNode.mSubNodesContent.Count));
            float angleInRad = GiWEngine.Math.Utilities.degreesToRadians(angleMin + ((angleMax - angleMin)/2.0f));
            int mCurrentLevel = subNode.mLevelContent;
            float distance = 0;
            for (int i = 0; i < mCurrentLevel + 1; ++i)
            {
                distance += (float)System.Math.Sqrt((float)mTree.mNodesPerLevelContent[i]);//((float)mTree.mNodesPerLevelContent[i] * (float)((float)mTree.mNodesPerLevelContent.Count * (float)mTree.mNodesPerLevelContent[i]);
            }
            
            subNode.mPositionContent = new GiWEngine.Math.Vector3((float)(Math.Cos(angleInRad) * /*(float)layer */ distance), (float)(Math.Sin(angleInRad) * /*(float)layer */ distance), 0.0f);
            subNode.mRectangleContent = null;
            subNode.mSphereSizeContent = 0.1f;
            if (!subNode.mIsLeafContent)
            {
                float amin = angleMin;
                float amax = amin + step;
                foreach (KeyValuePair<int, TreeNode> kvp in subNode.mSubNodesContent)
                {
                    computeSubNodeLayout(kvp.Value, amin, amax, layer + 1);
                    amin += step;
                    amax += step;
                }
            }
        }
        #endregion

        #region overridden Methods
        /**
         * Method for Computing the current Layout
         */
        public override Boolean computeLayout()
        {
            TreeNode root = mTree.mRootNodeContent;
            root.mPositionContent = new GiWEngine.Math.Vector3(0.0f, 0.0f, 0.0f);
            float angleMin = 0.0f;
            float step;
            float angleMax = step = 360.0f / root.mSubNodesContent.Count;
            root.mSphereSizeContent = 0.1f;
            root.mRectangleContent = null;
            foreach(KeyValuePair<int, TreeNode> kvp in root.mSubNodesContent)
            {
                computeSubNodeLayout(kvp.Value, angleMin, angleMax, 1);
                angleMin += step;
                angleMax += step;
            }
            return true;
        }


        /**
         * Method for generating the Displaylists for Tree
         */
        public override void generateTreeList()
        {
            mTreeList = Gl.glGenLists(1);
            Gl.glNewList(mTreeList, Gl.GL_COMPILE);
            Gl.glColor4f(0.9f, 0.9f, 0.9f, 0.9f);
            foreach (KeyValuePair<int, TreeNode> kvp in mTree.mRootNodeContent.mSubNodesContent)
            {
                Gl.glBegin(Gl.GL_LINES);
                Gl.glVertex3fv(mTree.mRootNodeContent.mPositionContent.toFloatArray());
                kvp.Value.genDisplayList();
            }
            Gl.glEndList();

            mTreeNodeList = Gl.glGenLists(1);
            Gl.glNewList(mTreeNodeList, Gl.GL_COMPILE);

            Gl.glShadeModel(Gl.GL_SMOOTH);

            Gl.glEnable(Gl.GL_LIGHTING);
            Gl.glEnable(Gl.GL_LIGHT0);
            float[] diffuseLight = { 0.8f, 0.8f, 0.8f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_DIFFUSE, diffuseLight);
            float[] specular = { 1.0f, 1.0f, 1.0f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_SPECULAR, specular);
            float[] ambient = { 1.0f, 1.0f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_AMBIENT, ambient);
            float[] lightpos = { 0.0f, 5.0f, -20.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_POSITION, lightpos);

            mTree.mRootNodeContent.draw();

            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glDisable(Gl.GL_LIGHT0);
            Gl.glEndList();
            
        }
        #endregion
    }
}
