using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Tao.OpenGl;

namespace InfoVis.Tree
{
    /**
     * Class representing an adjacency Pair
     * Provides the necessary functions to remove LCA, compute Bezier Splines
     */
    public class adjacencyPair
    {
        #region Member Variables
        public TreeNode start; /**< Start Node of AP */
        public TreeNode end; /**< End Node of AP */
        public List<TreeNode> mControlPoints; /**< Original ControlPoints of path */
        public TreeNode lca; /**< LCA */
        public List<GiWEngine.Math.Vector3> mFlattenedControlPoints; /**< Flattened ControlPoints of path */
        private int curveList; /**< Index in CurveList */
        private float[] controlPoints; /**< Positions of the ControlPoints of path */
        public static int mostControlPoints; /**< number of max Control Points in all Paths */
        //public List<TreeNode> fullTraversal;
        public List<GiWEngine.Math.Vector3> mCurvePoints; /**< Points of Curve */
        public Boolean mSelected; /**< If Curve is selected */
        #endregion

        #region Constructor
        /**
         * CTR of adjacencyPair
         */
        public adjacencyPair()
        {
            mostControlPoints = 0;
            mControlPoints = new List<TreeNode>();
            mFlattenedControlPoints = new List<GiWEngine.Math.Vector3>();
            //fullTraversal = new List<TreeNode>();
            mSelected = false;
        }
        #endregion

        #region Methods
        /**
         * computes the BinomialCoefficients for a given number of Control Points
         * @param numControlPoints number of controlPoints
         * @return List<int> the BinomialCoefficients
         */
        private List<int> computeBinomialCoefficients(int numControlPoints)
        {
            int n = numControlPoints;
            List<int> bC = new List<int>();
            for (int k = 0; k <= n; k++)
            {
                int tmp = 1;
                for (int j = n; j >= k + 1; j--)
                {
                    tmp *= j;
                }
                for (int j = n - k; j >= 2; j--)
                {
                    tmp /= j;
                }
                bC.Add(tmp);
            }
            return bC;
        }

        /**
         * Computes a BezierPoint for a given u
         * @param u interval on Curve
         * @param numControlpoints number of controlPoints
         * @param ctlrPtr List<Vector3> controlPoints
         * @param coefficients BinomialCoefficients
         * @returns Vector3 BezierPoint
         */
        private GiWEngine.Math.Vector3 computeBezierPoint(float u, int numControlPoints, List<GiWEngine.Math.Vector3> ctlrPts, List<int> coefficients)
        {
            int n = numControlPoints - 1;
            float bezBlendFcn;
            GiWEngine.Math.Vector3 bezPnt = new GiWEngine.Math.Vector3();

            for (int k = 0; k < numControlPoints; k++)
            {
                bezBlendFcn = coefficients[k] * (float)Math.Pow(u,k) * (float)Math.Pow(1 - u, n - k);
                bezPnt.x += ctlrPts[k].x * bezBlendFcn;
                bezPnt.y += ctlrPts[k].y * bezBlendFcn;
                bezPnt.z += ctlrPts[k].z * bezBlendFcn;
            }
            return bezPnt;
        }

        /**
         * builds DisplayList
         * @param varyingAlpha if curvelength changes alpha
         * @param colorDir if direction is color coded
         */
        public void buildDisplayList(Boolean varyingAlpha, Boolean colorDir)
        {
            controlPoints = new float[mFlattenedControlPoints.Count * 3];
            for (int i = 0; i < mFlattenedControlPoints.Count; i++)
            {
                controlPoints[3 * i] = mFlattenedControlPoints[i].toFloatArray()[0];
                controlPoints[(3 * i) + 1] = mFlattenedControlPoints[i].toFloatArray()[1];
                controlPoints[(3 * i) + 2] = mFlattenedControlPoints[i].toFloatArray()[2];
            }
            mCurvePoints = new List<GiWEngine.Math.Vector3>();
            curveList = Gl.glGenLists(1);
            Gl.glNewList(curveList, Gl.GL_COMPILE);
            //Gl.glMap1f(Gl.GL_MAP1_VERTEX_3, 0.0f, 1.0f, 3, mFlattenedControlPoints.Count, controlPoints);
            //Gl.glEnable(Gl.GL_MAP1_VERTEX_3);
            Gl.glBegin(Gl.GL_LINE_STRIP);
            int curveResolution = mFlattenedControlPoints.Count * 10;
            float alphavalue = (1.0f - ((float)mFlattenedControlPoints.Count / (float)mostControlPoints) + 0.3f);
            List<int> binCoeff = computeBinomialCoefficients(mFlattenedControlPoints.Count - 1);
            for (int i = 0; i <= curveResolution; i++)
            {
                if (varyingAlpha && colorDir)
                {
                    Gl.glColor4f(0.0f + (i * (1.0f / curveResolution)), 1.0f - (i * (1.0f / curveResolution)), 0.0f, alphavalue);
                }
                else if (!varyingAlpha && colorDir)
                {
                    Gl.glColor3f(0.0f + (i * (1.0f / curveResolution)), 1.0f - (i * (1.0f / curveResolution)), 0.0f);
                }
                else if (varyingAlpha && !colorDir)
                {
                    Gl.glColor4f(0.0f, 0.0f, 1.0f, alphavalue);
                }
                else if (!varyingAlpha && !colorDir)
                {
                    Gl.glColor3f(0.0f, 0.0f, 1.0f);
                }

                float u = (float)i / (float)curveResolution;
                GiWEngine.Math.Vector3 bezPoint = computeBezierPoint(u, mFlattenedControlPoints.Count, mFlattenedControlPoints, binCoeff);
                Gl.glVertex3fv(bezPoint.toFloatArray());
                mCurvePoints.Add(bezPoint);
                //Gl.glEvalCoord1f((float)i / (float)curveResolution);
            }
            Gl.glEnd();
            Gl.glDisable(Gl.GL_MAP1_VERTEX_3);
            Gl.glEndList();
        }

        /**
         * draws the current curve
         */
        public void drawDisplayList()
        {
            Gl.glCallList(curveList);
            /*Gl.glMap1f(Gl.GL_MAP1_VERTEX_3, 0.0f, 1.0f, 3, mFlattenedControlPoints.Count, controlPoints);
            Gl.glEnable(Gl.GL_MAP1_VERTEX_3);
            Gl.glBegin(Gl.GL_LINE_STRIP);
            int curveResolution = mFlattenedControlPoints.Count * 10;
            for (int i = 0; i <= curveResolution; i++)
                Gl.glEvalCoord1f((float)i / (float)curveResolution);
            Gl.glEnd();*/
        }

        /**
         * Intersects a given Line with the current Curve
         * @param m_StartPoint starting Point of line
         * @param m_End end Point of Line
         * @returns if intersect
         */
        public Boolean intersectWithLine(GiWEngine.Math.Vector3 m_StartPoint, GiWEngine.Math.Vector3 m_End)
        {
            for (int i = 0; i < mCurvePoints.Count - 2; ++i)
            {
                GiWEngine.Math.Vector3 mP1 = mCurvePoints[i];
                GiWEngine.Math.Vector3 mP2 = mCurvePoints[i + 1];

                float ua = ((m_End.x - m_StartPoint.x)*(mP1.y - m_StartPoint.y) - (m_End.y - m_StartPoint.y)*(mP1.x - m_StartPoint.x)) /
                    ((m_End.y - m_StartPoint.y)*(mP2.x - mP1.x) - (m_End.x - m_StartPoint.x)*(mP2.y - mP1.y));
                float ub = ((mP2.x - mP1.x)*(mP1.y - m_StartPoint.y) - (mP2.y - mP1.y)*(mP1.x - m_StartPoint.x)) /
                    ((m_End.y - m_StartPoint.y) * (mP2.x - mP1.x) - (m_End.x - m_StartPoint.x) * (mP2.y - mP1.y));
                if (((0.0f < ua) && (ua < 1.0f)) && ((0.0f < ub) && (ub < 1.0f)))
                {
                    mSelected = true;
                    return true;
                }
            }
            mSelected = false;
            return false;
        }
        #endregion
    }

    /**
     * Class representing a Tree Stucture
     */
    public class Tree
    {
        #region Member Variables
        private TreeNode mRootNode;/**< Root Node of AP */
        private List<adjacencyPair> mAllAdjacencies; /**< List of all Adjacencies */
        private List<int> mNodesPerLevel; /**< How many Nodes per Level, represented via index */
        #endregion

        #region Constructor
        /**
         * CTR of Tree
         * @param m_RootNode rootNode of Tree
         * @param m_AllAdjacencies List of all Adjacencies
         */
        public Tree(TreeNode m_RootNode, List<adjacencyPair> m_AllAdjacencies)
        {
            mRootNode = m_RootNode;
            mAllAdjacencies = m_AllAdjacencies;
            mNodesPerLevel = new List<int>();
        }
        #endregion

        #region getter
        public List<int> mNodesPerLevelContent
        {
            get
            {
                return mNodesPerLevel;
            }
        }

        public TreeNode mRootNodeContent
        {
            get
            {
                return mRootNode;
            }
        }

        public List<adjacencyPair> mAllAdjacenciesContent
        {
            get
            {
                return mAllAdjacencies;
            }
            set
            {
                mAllAdjacencies = value;
            }
        }
        #endregion

        #region Methods
        /**
         * Adds Node to the Tree
         * @param m_Id int to which note to add
         * @param m_SubNode The Node(s) to add
         */
        public void addNode(int m_Id, TreeNode m_SubNode)
        {
            if (mRootNode.mIdContent == m_Id)
            {
                mRootNode.addSubNode(m_SubNode);
            }
            else
            {
                mRootNode.addIntoSubNode(m_Id, m_SubNode);
            }
        }

        /**
         * flattenes the control Points in each Adjacency Pair and rebuilds display Lists
         * @param beta float value [0,1] which is used to flatten
         * @param varyingAlpha if the alpha value should be varied
         * @param colorDir if the color codes the direction
         */
        public void flattenControlPoints(float beta, Boolean varyingAlpha, Boolean colorDir)
        {
            foreach (adjacencyPair ap in mAllAdjacencies)
            {
                adjacencyPair.mostControlPoints = 0;
                GiWEngine.Math.Vector3 P0 = ap.start.mPositionContent;
                GiWEngine.Math.Vector3 PN_1 = ap.end.mPositionContent;
                int N = ap.mControlPoints.Count;
                if (N > 1)
                {
                    ap.mFlattenedControlPoints.Clear();
                    for (int i = 0; i < ap.mControlPoints.Count; ++i)
                    {
                        GiWEngine.Math.Vector3 Pi = ap.mControlPoints[i].mPositionContent;
                        GiWEngine.Math.Vector3 P_i = beta * Pi + (1 - beta) *
                            (P0 + (float)(i / (N - 1)) * (PN_1 - P0));
                        ap.mFlattenedControlPoints.Add(P_i);
                    }
                    if (ap.mFlattenedControlPoints.Count > adjacencyPair.mostControlPoints)
                    {
                        adjacencyPair.mostControlPoints = ap.mFlattenedControlPoints.Count;
                    }
                }
            }
            mAllAdjacencies.Sort(delegate(adjacencyPair a1, adjacencyPair a2) { return a1.mFlattenedControlPoints.Count.CompareTo(a2.mFlattenedControlPoints.Count); });
            foreach (adjacencyPair ap in mAllAdjacencies)
            {
                ap.buildDisplayList(varyingAlpha, colorDir);
            }
        }

        /**
         * Counts how many Nodes Per level
         */
        public void enumerateLevels()
        {
            mRootNode.mLevelContent = 0;
            mNodesPerLevel.Add(1);
            foreach (KeyValuePair<int, TreeNode> kvp in mRootNode.mSubNodesContent)
            {
                kvp.Value.enumerateSubNodes(1, mNodesPerLevel);
            }
        }
        #endregion

        #region Static Methods
        /**
         * Loads a Tree from given String
         * @param fileURI string containing the Filename of heb file
         * @returns Tree Object
         */
        public static Tree loadFromFile(string fileURI)
        {
            FileStream fs = new FileStream(fileURI, FileMode.Open, FileAccess.Read, FileShare.None);
            StreamReader sr = new StreamReader(fs);

            Dictionary<int, TreeNode> readNodes = new Dictionary<int, TreeNode>();
            List<adjacencyPair> allAdjacencies = new List<adjacencyPair>();
            string readline;
            //reads all nodes
            while (!(readline = sr.ReadLine()).Equals("inheritance"))
            {
                string[] tmp = readline.Split(' ');
                int id = Int32.Parse(tmp[0]);
                //Boolean leaf = Boolean.Parse(tmp[2]);
                readNodes.Add(id, new TreeNode(id, tmp[1]));
            }

            //reads the inheritance
            while (!(readline = sr.ReadLine()).Equals("adjacency")) 
            {
                string[] tmp = readline.Split(' ');
                int ancestor = Int32.Parse(tmp[0]);
                int descendant = Int32.Parse(tmp[1]);
                TreeNode tmpNode;
                readNodes.TryGetValue(ancestor, out tmpNode);
                if (tmpNode != null)
                {
                    TreeNode descendantNode;
                    readNodes.TryGetValue(descendant, out descendantNode);
                    if (descendantNode != null)
                    {
                        tmpNode.addSubNode(descendantNode);
                    }
                }
            }

            //reads adjacencies
            while (!(readline = sr.ReadLine()).Equals(""))
            {
                string[] tmp = readline.Split(' ');
                int start = Int32.Parse(tmp[0]);
                int end = Int32.Parse(tmp[1]);
                adjacencyPair tmpPair = new adjacencyPair();
                TreeNode startN, endN;
                if ((readNodes.TryGetValue(start, out startN)) && (readNodes.TryGetValue(end, out endN)))
                {
                    startN.addAdjacent(endN);
                    tmpPair.start = startN;
                    tmpPair.end = endN;
                    allAdjacencies.Add(tmpPair);
                }
            }

            sr.Close();
            fs.Close();
            TreeNode root;
            readNodes.TryGetValue(0, out root);

            return new Tree(root, allAdjacencies);
        }
        #endregion
    }
}
