using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;

namespace InfoVis.Tree
{
    /**
     * Abstract Class Providing all necessary Member Variables and Functions used by a TreeLayout
     */
    public abstract class TreeLayout
    {
        #region Member Variables
        protected Tree mTree;/**< Tree used in this Layout */
        protected int mTreeList; /**< value where Displaylist of TreeStructure is stored */
        protected int mTreeNodeList; /**< value where Displaylist of Nodes is stored */
        protected TreeNode mSelectedNode; /**< current selected Node */
        protected GiWEngine.Math.Vector3 mStartSel; /**< Starting Point of Line Selection*/
        protected GiWEngine.Math.Vector3 mEndSel; /**< End Points of Line Selection*/
        protected Boolean mLinesSelected; /**< If Lines are selected*/
        #endregion

        #region Constructor
        /*public TreeLayout(ref Tree m_Tree)
        {
            mTree = Tree;
        }*/
        #endregion

        #region Methods
        /**
         * Method for Computing the current Layout
         */
        public abstract Boolean computeLayout();
        /**
         * Method for generating the Displaylists for Tree
         */
        public abstract void generateTreeList();

        /**
         * Draws the Current layout
         * @param alphaPass if it should draw the alpha Pass
         */
        public void draw(Boolean alphaPass)
        {
            Gl.glEnable(Gl.GL_BLEND);
            if (alphaPass)
            {
                Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);
            }
            else
            {
                Gl.glBlendEquationEXT(Gl.GL_MIN_EXT);
            }
            if (mSelectedNode != null)
            {
                foreach (adjacencyPair ap in mTree.mAllAdjacenciesContent)
                {
                    if (ap.mControlPoints.Contains(mSelectedNode))
                    {
                        ap.drawDisplayList();
                    }
                    else if (ap.lca != null)
                    {
                        if (ap.lca.mIdContent == mSelectedNode.mIdContent)
                        {
                            ap.drawDisplayList();
                        }
                    }
                }
            }
            else if((mStartSel != null) && (mEndSel != null) && !mLinesSelected)
            {
                foreach (adjacencyPair ap in mTree.mAllAdjacenciesContent)
                {
                    if(ap.intersectWithLine(mStartSel, mEndSel))
                    {
                        ap.drawDisplayList();
                    }
                }
                mLinesSelected = true;
            }
            else if ((mStartSel != null) && (mEndSel != null) && mLinesSelected)
            {
                foreach (adjacencyPair ap in mTree.mAllAdjacenciesContent)
                {
                    if (ap.mSelected)
                    {
                        ap.drawDisplayList();
                    }
                }
                mLinesSelected = true;
            }
            else
            {
                foreach (adjacencyPair ap in mTree.mAllAdjacenciesContent)
                {
                    ap.drawDisplayList();
                }
            }
            Gl.glDisable(Gl.GL_BLEND);
            Gl.glCallList(mTreeList);


            /*Gl.glShadeModel(Gl.GL_SMOOTH);

            Gl.glEnable(Gl.GL_LIGHTING);
            Gl.glEnable(Gl.GL_LIGHT0);
            float[] diffuseLight = { 0.8f, 0.8f, 0.8f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_DIFFUSE, diffuseLight);
            float[] specular = { 1.0f, 1.0f, 1.0f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_SPECULAR, specular);
            float[] ambient = { 1.0f, 1.0f, 1.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_AMBIENT, ambient);
            float[] lightpos = { 0.0f, 5.0f, -20.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_POSITION, lightpos);

            mTree.mRootNodeContent.draw();

            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glDisable(Gl.GL_LIGHT0);*/

            Gl.glCallList(mTreeNodeList);
        }

        /**
         * Method for finding and Removing the LCA in Two Lists
         * @param m_start List of Nodes beginning at the Startpoint to Root
         * @param m_start List of Nodes beginning at the Endpoint to Root
         */
        protected void removeUntilLCA(List<TreeNode> m_start, List<TreeNode> m_end)
        {
            int startLength = m_start.Count - 1;
            int endLength = m_end.Count - 1;

            while ((startLength != 0) && (endLength != 0) && (m_start[startLength - 1] == m_end[endLength - 1]))
            {
                if ((m_start[startLength] == m_end[endLength]) &&
                    m_start[startLength - 1] == m_end[endLength - 1])
                {
                    m_start.RemoveAt(startLength);
                    startLength--;

                    m_end.RemoveAt(endLength);
                    endLength--;
                }
            }
        }

        /**
         * Method which finds and generates the Controlpoints building a Bezier curve connecting two Nodes
         */
        public void generateAdjacencyControlPoints()
        {
            for (int i = 0; i < mTree.mAllAdjacenciesContent.Count; ++i)
            {
                List<TreeNode> toStart = traverseForNode(mTree.mAllAdjacenciesContent[i].start.mIdContent);
                List<TreeNode> toEnd = traverseForNode(mTree.mAllAdjacenciesContent[i].end.mIdContent);
                removeUntilLCA(toStart, toEnd);
                //mTree.mAllAdjacenciesContent[i].fullTraversal.AddRange(toStart);
                //mTree.mAllAdjacenciesContent[i].fullTraversal.AddRange(toEnd);
                mTree.mAllAdjacenciesContent[i].lca = toStart[toStart.Count - 1];
                if ((toStart.Count - 1 + toEnd.Count - 1) > 3) //optimization of paper where LCA is removed if more than 3 control points
                {
                    toStart.RemoveAt(toStart.Count - 1);
                    toEnd.RemoveAt(toEnd.Count - 1);

                }
                else
                {
                    toEnd.RemoveAt(toEnd.Count - 1); //removes double LCA
                }
                toEnd.Reverse();
                toStart.AddRange(toEnd);
                adjacencyPair tmp = new adjacencyPair();
                tmp.start = mTree.mAllAdjacenciesContent[i].start;
                tmp.end = mTree.mAllAdjacenciesContent[i].end;
                tmp.mControlPoints = toStart;
                tmp.lca = mTree.mAllAdjacenciesContent[i].lca;
                //tmp.fullTraversal = mTree.mAllAdjacenciesContent[i].fullTraversal;
                mTree.mAllAdjacenciesContent[i] = tmp;
            }
        }

        /**
         * Traversed the Tree for a given Node
         * @param m_Id Node Id of node to be found
         * @return List of TreeNodes representing the traversal
         */
        public List<TreeNode> traverseForNode(int m_Id)
        {
            List<TreeNode> traversal = new List<TreeNode>();
            mTree.mRootNodeContent.searchNode(traversal, m_Id);
            return traversal;
        }

        /**
         * updates the boundling Strenght and flattens the Controll Points according to it
         * @param beta the boundling strength
         * @param varyingAlpha if the Alpha depents on the amount of control Points
         * @param colorDir if direction is color coded
         */
        public void updateBoundlingStrength(float beta, Boolean varyingAlpha, Boolean colorDir)
        {
            mTree.flattenControlPoints(beta, varyingAlpha, colorDir);
        }

        /**
         * picks a Node from the Tree
         * @param p_near Starting point of ray
         * @param m_nner Point giving the direction of the ray
         * @returns string with name of picked node
         */
        public string rayPick(GiWEngine.Math.Vector3 p_near, GiWEngine.Math.Vector3 p_nnear)
        {
            //constructs ray
            GiWEngine.Math.Vector3 ray = p_nnear - p_near;
            ray.normalize();

            TreeNode tmp = mTree.mRootNodeContent.rayPick(p_near, ray);
            if (tmp == null)
            {
                return "";
            }
            else
            {
                return tmp.mIdContent + ", " + tmp.mNameContent;
            }
        }

        /**
         * selects a Node from the Tree
         * @param p_near Starting point of ray
         * @param m_nner Point giving the direction of the ray
         * @returns Selected Node
         */
        public TreeNode select(GiWEngine.Math.Vector3 p_near, GiWEngine.Math.Vector3 p_nnear)
        {
            //constructs ray
            GiWEngine.Math.Vector3 ray = p_nnear - p_near;
            ray.normalize();

            return mTree.mRootNodeContent.rayPick(p_near, ray);
        }

        /**
         * Sets the current Selected node of Tree
         */
        public void setSelectedNode(TreeNode m_selNode)
        {
            mSelectedNode = m_selNode;
        }

        /**
         * Tells the Layout if some Curves are Selected with a given Line
         * @param m_StartPoint start point of line
         * @param m_EndPoint end point of line
         */
        public void selectWithLine(GiWEngine.Math.Vector3 m_StartPoint, GiWEngine.Math.Vector3 m_EndPoint)
        {
            mStartSel = m_StartPoint;
            mEndSel = m_EndPoint;
            mLinesSelected = false;
        }
        #endregion
    }
}
