using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;

namespace InfoVis.Tree
{
    /**
     * Class representing a TreeNode
     */
    public class TreeNode
    {
        #region Member Variables
        private int mId; /**< node Id */
        private string mName; /**< name of Node */
        private GiWEngine.Math.Vector3 mPosition; /**< Position in Current Layout */
        private Boolean mIsLeaf; /**< if it has Subnodes */
        private Dictionary<int, TreeNode> mSubNodes; /**< Subnodes */
        private Dictionary<int, TreeNode> mAdjacents; /**< Nodes Adjacent to this node */
        private int mLevel; /**< absolute level in tree */
        private List<GiWEngine.Math.Vector3> mRectangle; /**< Edge Points of Rectangle in Treemap */
        private float mSphereSize; /**< Size of Visual Node */ 
        #endregion

        #region Constructor
        /**
         * CTR of TreeNode
         * @param m_Id Id of Node
         * @param m_Name Name of Node
         * @param m_IsLeaf if it is a leafnode
         */
        public TreeNode(int m_Id, string m_Name, Boolean m_IsLeaf)
        {
            mId = m_Id;
            mName = m_Name;
            mIsLeaf = m_IsLeaf;
            if (!mIsLeaf)
            {
                mSubNodes = new Dictionary<int, TreeNode>();
            }          
            mAdjacents = new Dictionary<int, TreeNode>();
            mSphereSize = 0.1f;
        }

        /**
         * CTR of TreeNode
         * @param m_Id Id of Node
         * @param m_Name Name of Node
         */
        public TreeNode(int m_Id, string m_Name)
        {
            mId = m_Id;
            mName = m_Name;
            mIsLeaf = true;
            mSubNodes = new Dictionary<int, TreeNode>();
            mAdjacents = new Dictionary<int, TreeNode>();
            mSphereSize = 0.1f;
        }
        #endregion

        #region getter /setter

        public List<GiWEngine.Math.Vector3> mRectangleContent
        {
            get
            {
                return mRectangle;
            }
            set
            {
                mRectangle = value;
            }
        }

        public float mSphereSizeContent
        {
            get
            {
                return mSphereSize;
            }
            set
            {
                mSphereSize = value;
            }
        }

        public int mLevelContent
        {
            get
            {
                return mLevel;
            }
            set
            {
                mLevel = value;
            }
        }

        public int mIdContent
        {
            get
            {
                return mId;
            }
        }

        public string mNameContent
        {
            get
            {
                return mName;
            }
        }

        public GiWEngine.Math.Vector3 mPositionContent
        {
            get
            {
                return mPosition;
            }
            set
            {
                mPosition = value;
            }
        }

        public Boolean mIsLeafContent
        {
            get
            {
                return mIsLeaf;
            }
        }

        public Dictionary<int, TreeNode> mSubNodesContent
        {
            get
            {
                return mSubNodes;
            }
        }

        public Dictionary<int, TreeNode> mAdjacentsContent
        {
            get
            {
                return mAdjacents;
            }
        }
        #endregion

        #region Methods
        /**
         * Searches for an Node
         * @param m_traversal list of already traversed Nodes
         * @param m_Id id of node to be found
         * @returns if it is found
         */
        public Boolean searchNode(List<TreeNode> m_traversal, int m_Id)
        {
            if (this.mId == m_Id)
            {
                m_traversal.Add(this);
                return true;
            }
            else
            {
                if (mIsLeaf)
                {
                    return false;
                }

                foreach (TreeNode tn in mSubNodes.Values)
                {
                    if(tn.searchNode(m_traversal, m_Id))
                    {
                        if (!m_traversal.Contains(this))
                        {
                            m_traversal.Add(this);
                        }
                        return true;
                    }
                }
                return false;
            }
        }

        /**
         * Adds a Node Adjacent to this
         * @param m_Node Adjacent Node
         */
        public void addAdjacent(TreeNode m_Node)
        {
            if (!mAdjacents.ContainsKey(m_Node.mIdContent))
            {
                mAdjacents.Add(m_Node.mIdContent, m_Node);
            }
        }

        /**
         * returns a Subnode with given Index
         * @param index index of Subnode
         * @returns Treenode with index
         */
        public TreeNode getSubNodeFromIndex(int index)
        {
            TreeNode tmp = null;
            if (mIsLeaf)
            {
                if (mSubNodes.Count - 1 < index)
                {
                    tmp = null;
                }
                else
                {
                    tmp = mSubNodes[index];
                }
            }
            return tmp;
        }

        /**
         * Adds a Subnode to this node
         * @param m_SubNode TreeNode to add
         */
        public void addSubNode(TreeNode m_SubNode)
        {
            if ((mSubNodes.Count == 0) || !mIsLeaf)
            {
                mIsLeaf = false;
            }

            mSubNodes.Add(m_SubNode.mIdContent, m_SubNode);
        }

        /**
         * Adds a Node into a given Subnode
         * @param m_Id where the node should be added
         * @param m_SubNode Node to be added
         */
        public void addIntoSubNode(int m_Id, TreeNode m_SubNode)
        {
            if (mSubNodes.ContainsKey(m_Id))
            {
                mSubNodes[m_Id].addSubNode(m_SubNode);
            }
            else
            {
                foreach (TreeNode tn in mSubNodes.Values)
                {
                    tn.addIntoSubNode(m_Id, m_SubNode);
                }
            }
        }

        /**
         * Searches and returns a Subnode
         * @param m_Id NodeId to be searched for
         * @returns TreeNode
         */
        public TreeNode getSubNodeFromId(int m_Id)
        {
            TreeNode tmp;
            if(mSubNodes.TryGetValue(m_Id, out tmp))
            {
                return tmp;
            }
            else
            {
                foreach(TreeNode tn in mSubNodes.Values)
                {
                    tmp = tn.getSubNodeFromId(m_Id);
                    if(tmp != null)
                    {
                        return tmp;
                    }
                }
            }
            return null;
        }
           
        /**
         * draws the current Node and its subnodes
         */
        public void draw()
        {
            #region Draw Sphere
            bool drawWireFrame = false;
            int drawStyle = drawWireFrame ? Glu.GLU_LINE : Glu.GLU_FILL;
            double radius = (double)mSphereSize;
            // From NeHe lesson 18.
            Gl.glPushMatrix();
            Glu.GLUquadric quadric = Glu.gluNewQuadric();
            try
            {
                Glu.gluQuadricDrawStyle(quadric, drawStyle);
                Glu.gluQuadricNormals(quadric, Glu.GLU_SMOOTH);
                float[] tmp = mPosition.toFloatArray();
                Gl.glTranslatef(tmp[0], tmp[1], tmp[2]);
                Glu.gluSphere(quadric, radius, 40, 40);
            }
            finally
            {
                Glu.gluDeleteQuadric(quadric);
            }
            Gl.glPopMatrix();
            #endregion Draw Sphere

            if (!mIsLeaf)
            {
                foreach (KeyValuePair<int, TreeNode> kvp in mSubNodes)
                {
                    kvp.Value.draw();
                }
            }
        }

        /**
         * Picks a Node with a given Ray
         * @param p0 Starting Point of Ray
         * @param ray Direction of ray
         * @returns TreeNode hit by ray
         */
        public TreeNode rayPick(GiWEngine.Math.Vector3 p0, GiWEngine.Math.Vector3 ray)
        {
            //Collide with itself
            GiWEngine.Math.Vector3 M = p0 - mPosition;
            float m2 = M * M;
            float r2 = (float)Math.Pow(((double)mSphereSize * 1.5d) ,2);
            float d = (float)Math.Pow(ray * M,2);
            float sub = m2 - r2;
            d = d - sub;
            if (d >= 0)
            {
                return this;
            }
            else
            {
                //if no collision collide with subnodes
                if (mIsLeaf)
                {
                    return null;
                }
                else
                {
                    foreach (KeyValuePair<int, TreeNode> kvp in mSubNodes)
                    {
                        TreeNode tmp = kvp.Value.rayPick(p0, ray);
                        if (tmp != null)
                        {
                            return tmp;
                        }
                    }
                    return null;
                }
            }
            

        }

        /**
         * generates all Display Lists
         */
        public void genDisplayList()
        {
            if (mRectangle != null)
            {
                Gl.glBegin(Gl.GL_LINE_LOOP);
                foreach(GiWEngine.Math.Vector3 pos in mRectangle)
                {
                    Gl.glVertex3fv(pos.toFloatArray());
                }
                Gl.glEnd();
                if (!mIsLeaf)
                {
                    foreach (KeyValuePair<int, TreeNode> kvp in mSubNodes)
                    {
                        kvp.Value.genDisplayList();
                    }
                }
            }
            else
            {
                Gl.glVertex3fv(mPosition.toFloatArray());
                Gl.glEnd();
                if (!mIsLeaf)
                {
                    foreach (KeyValuePair<int, TreeNode> kvp in mSubNodes)
                    {
                        Gl.glBegin(Gl.GL_LINES);
                        Gl.glVertex3fv(mPosition.toFloatArray());
                        kvp.Value.genDisplayList();
                    }
                }
            }
        }

        /**
         * enumerates the levels of the subnodes
         * @param m_level current Level in Subnode
         * @param m_levelcount nodes in this Level
         */
        public void enumerateSubNodes(int m_level, List<int> m_levelcount)
        {
            mLevel = m_level;
            if (m_levelcount.Count == m_level)
            {
                m_levelcount.Add(1);
            }
            else
            {
                m_levelcount[mLevel]++;
            }
            if (!mIsLeaf)
            {
                foreach (KeyValuePair<int, TreeNode> kvp in mSubNodesContent)
                {
                    kvp.Value.enumerateSubNodes(m_level+1, m_levelcount);
                }
            }
        }

        /**
         * Sets the RectangeCoordinates of Node in TreeMapLayout
         * @param m_Rectangle List of Edgepoints
         */
        public void setRectangeCoordinates(List<GiWEngine.Math.Vector3> m_Rectangle)
        {
            mRectangle = m_Rectangle;
        }
        #endregion
    }
}
